import argparse

import annofabcli.stat_visualization.write_performance_rating_csv


def parse_args(parser: argparse.ArgumentParser):

    subparsers = parser.add_subparsers(dest="subcommand_name")

    # サブコマンドの定義
    annofabcli.stat_visualization.write_performance_rating_csv.add_parser(subparsers)


def add_parser(subparsers: argparse._SubParsersAction):
    subcommand_name = "stat_visualization"
    subcommand_help = "`annofabcli statistics visualization`コマンドの出力結果を加工するサブコマンド"
    description = "`annofabcli statistics visualization`コマンドの出力結果を加工するサブコマンド"

    parser = annofabcli.common.cli.add_parser(
        subparsers, subcommand_name, subcommand_help, description, is_subcommand=False
    )
    parse_args(parser)



import argparse
import logging
from typing import Any, Dict, List, Optional

import annofabapi
import pandas
from annofabapi.models import Task

import annofabcli
from annofabcli import AnnofabApiFacade
from annofabcli.common.cli import AbstractCommandLineInterface, ArgumentParser, build_annofabapi_resource_and_login
from annofabcli.common.enums import FormatArgument
from annofabcli.common.utils import get_columns_with_priority
from annofabcli.common.visualize import AddProps

logger = logging.getLogger(__name__)


class ListTasksMain:
    def __init__(self, service: annofabapi.Resource, project_id: str):
        self.service = service
        self.facade = AnnofabApiFacade(service)
        self.project_id = project_id
        self.visualize = AddProps(self.service, project_id)

    @annofabcli.utils.allow_404_error
    def get_task(self, project_id: str, task_id: str) -> Task:
        task, _ = self.service.api.get_task(project_id, task_id)
        return task

    def get_task_list_from_task_id(self, project_id: str, task_id_list: List[str]) -> List[Task]:
        task_list = []
        logger.debug(f"{len(task_id_list)}件のタスクを取得します。")
        for index, task_id in enumerate(task_id_list):
            if (index + 1) % 100 == 0:
                logger.debug(f"{index+1} 件のタスクを取得します。")

            task = self.get_task(project_id, task_id)
            if task is not None:
                task_list.append(self.visualize.add_properties_to_task(task))
            else:
                logger.warning(f"タスク '{task_id}' は見つかりませんでした。")

        return task_list

    def _modify_task_query(self, project_id: str, task_query: Dict[str, Any]) -> Dict[str, Any]:
        """
        タスク検索クエリを修正する。
        ``user_id`` から ``account_id`` に変換する。
        ``previous_user_id`` から ``previcous_account_id`` に変換する。
        ``page`` , ``limit``を削除」する

        Args:
            task_query: タスク検索クエリ（変更される）

        Returns:
            修正したタスク検索クエリ

        """

        def remove_key(arg_key: str):
            if arg_key in task_query:
                logger.info(f"タスク検索クエリから、`{arg_key}`　キーを削除しました。")
                task_query.pop(arg_key)

        remove_key("page")
        remove_key("limit")

        if "user_id" in task_query:
            user_id = task_query["user_id"]
            account_id = self.facade.get_account_id_from_user_id(project_id, user_id)
            if account_id is not None:
                task_query["account_id"] = account_id
            else:
                logger.warning(f"タスク検索クエリに含まれている user_id: {user_id} のユーザが見つかりませんでした。")

        if "previous_user_id" in task_query:
            previous_user_id = task_query["previous_user_id"]
            previous_account_id = self.facade.get_account_id_from_user_id(project_id, previous_user_id)
            if previous_account_id is not None:
                task_query["previous_account_id"] = previous_account_id
            else:
                logger.warning(f"タスク検索クエリに含まれている previous_user_id: {previous_user_id} のユーザが見つかりませんでした。")

        return task_query

    def get_task_list_with_api(
        self, project_id: str, task_query: Optional[Dict[str, Any]] = None, user_id_list: Optional[List[str]] = None
    ) -> List[Task]:
        """
        タスク一覧を取得する。

        Args:
            project_id:
            task_id_list:

        Returns:
            対象の検査コメント一覧
        """
        if task_query is not None:
            task_query = self._modify_task_query(project_id, task_query)
        else:
            task_query = {}

        if user_id_list is None:
            logger.debug(f"task_query: {task_query}")
            tasks = self.service.wrapper.get_all_tasks(project_id, query_params=task_query)
            if len(tasks) == 10000:
                logger.warning("タスク一覧は10,000件で打ち切られている可能性があります。")

        else:
            tasks = []
            for user_id in user_id_list:
                task_query["user_id"] = user_id
                task_query = self._modify_task_query(project_id, task_query)
                logger.debug(f"task_query: {task_query}")
                sub_tasks = self.service.wrapper.get_all_tasks(project_id, query_params=task_query)
                if len(sub_tasks) == 10000:
                    logger.warning(f"user_id={user_id}で絞り込んだタスク一覧は10,000件で打ち切られている可能性があります。")
                tasks.extend(sub_tasks)

        return [self.visualize.add_properties_to_task(e) for e in tasks]

    def get_task_list(
        self,
        project_id: str,
        task_id_list: Optional[List[str]] = None,
        task_query: Optional[Dict[str, Any]] = None,
        user_id_list: Optional[List[str]] = None,
    ) -> List[Task]:
        """


        Args:
            project_id: 対象のproject_id
            task_id_list: 対象のタスクのtask_id
            task_query: タスク検索クエリ
            user_id_list:

        Returns:

        """
        if task_id_list is not None:
            task_list = self.get_task_list_from_task_id(project_id, task_id_list=task_id_list)
        else:
            task_list = self.get_task_list_with_api(project_id, task_query=task_query, user_id_list=user_id_list)

        return task_list


class ListTasks(AbstractCommandLineInterface):
    """
    タスクの一覧を表示する
    """

    def __init__(self, service: annofabapi.Resource, facade: AnnofabApiFacade, args: argparse.Namespace):
        super().__init__(service, facade, args)
        self.visualize = AddProps(self.service, args.project_id)

    PRIOR_COLUMNS = [
        "project_id",
        "task_id",
        "phase",
        "phase_stage",
        "status",
        "started_datetime",
        "updated_datetime",
        "user_id",
        "username",
        "worktime_hour",
        "metadata",
        "number_of_rejections_by_inspection",
        "number_of_rejections_by_acceptance",
    ]

    def main(self):
        args = self.args

        task_id_list = annofabcli.common.cli.get_list_from_args(args.task_id)
        if len(task_id_list) == 0:
            task_id_list = None

        user_id_list = annofabcli.common.cli.get_list_from_args(args.user_id)
        if len(user_id_list) == 0:
            user_id_list = None

        task_query = annofabcli.common.cli.get_json_from_args(args.task_query)

        project_id = args.project_id
        super().validate_project(project_id, project_member_roles=None)

        main_obj = ListTasksMain(self.service, project_id=project_id)
        task_list = main_obj.get_task_list(
            project_id=project_id,
            task_id_list=task_id_list,
            task_query=task_query,
            user_id_list=user_id_list,
        )
        logger.debug(f"タスク一覧の件数: {len(task_list)}")

        if len(task_list) > 0:
            if self.str_format == FormatArgument.CSV.value:
                df = pandas.DataFrame(task_list)
                columns = get_columns_with_priority(df, prior_columns=self.PRIOR_COLUMNS)
                self.print_csv(df[columns])
            else:
                self.print_according_to_format(task_list)
        else:
            logger.info(f"タスク一覧の件数が0件のため、出力しません。")


def main(args):
    service = build_annofabapi_resource_and_login(args)
    facade = AnnofabApiFacade(service)
    ListTasks(service, facade, args).main()


def parse_args(parser: argparse.ArgumentParser):
    argument_parser = ArgumentParser(parser)

    argument_parser.add_project_id()

    query_group = parser.add_mutually_exclusive_group()

    # タスク検索クエリ
    query_group.add_argument(
        "-tq",
        "--task_query",
        type=str,
        help="タスクの検索クエリをJSON形式で指定します。指定しない場合は、すべてのタスクを取得します。"
        "`file://`を先頭に付けると、JSON形式のファイルを指定できます。"
        "クエリのフォーマットは、[getTasks API](https://annofab.com/docs/api/#operation/getTasks)のクエリパラメータと同じです。"
        "さらに追加で、`user_id`, `previous_user_id` キーも指定できます。"
        "ただし `page`, `limit`キーは指定できません。",
    )

    query_group.add_argument(
        "-t",
        "--task_id",
        type=str,
        nargs="+",
        help="対象のタスクのtask_idを指定します。`--task_query`引数とは同時に指定できません。" "`file://`を先頭に付けると、task_idの一覧が記載されたファイルを指定できます。",
    )

    parser.add_argument(
        "-u",
        "--user_id",
        type=str,
        nargs="+",
        help="絞り込み対象である担当者のuser_idを指定します。" "`file://`を先頭に付けると、task_idの一覧が記載されたファイルを指定できます。",
    )

    argument_parser.add_format(
        choices=[FormatArgument.CSV, FormatArgument.JSON, FormatArgument.PRETTY_JSON, FormatArgument.TASK_ID_LIST],
        default=FormatArgument.CSV,
    )
    argument_parser.add_output()
    argument_parser.add_csv_format()

    argument_parser.add_query()
    parser.set_defaults(subcommand_func=main)


def add_parser(subparsers: argparse._SubParsersAction):
    subcommand_name = "list"
    subcommand_help = "タスク一覧を出力します。"
    description = "タスク一覧を出力します。"

    parser = annofabcli.common.cli.add_parser(subparsers, subcommand_name, subcommand_help, description)
    parse_args(parser)



import argparse
import logging
from typing import List

import annofabapi
import requests
from annofabapi.models import ProjectMemberStatus
from more_itertools import first_true

import annofabcli
from annofabcli import AnnofabApiFacade
from annofabcli.common.cli import AbstractCommandLineInterface, build_annofabapi_resource_and_login

logger = logging.getLogger(__name__)


class DropProjectMembersMain:
    def __init__(self, service: annofabapi.Resource):
        self.service = service
        self.facade = AnnofabApiFacade(service)

    def drop_project_members(self, project_id: str, user_id_list: List[str]):
        """
        複数ユーザをプロジェクトメンバから脱退させる

        Args:
            project_id:
            user_id_list:

        """
        dest_project_members = self.service.wrapper.get_all_project_members(project_id)

        def get_project_member(user_id: str):
            return first_true(dest_project_members, pred=lambda e: e["user_id"] == user_id)

        project_title = self.facade.get_project_title(project_id)

        # プロジェクトメンバを追加/更新する
        for user_id in user_id_list:
            dest_member = get_project_member(user_id)
            if dest_member is None or dest_member["member_status"] == ProjectMemberStatus.INACTIVE.value:
                logger.debug(f"{project_title}({project_id}) のプロジェクトメンバに、{user_id} は存在しないので、スキップします。")
                continue

            request_body = {
                "member_status": ProjectMemberStatus.INACTIVE.value,
                "member_role": dest_member["member_role"],
                "last_updated_datetime": dest_member["updated_datetime"],
            }
            try:
                self.service.api.put_project_member(project_id, user_id, request_body=request_body)
                logger.debug(f"{project_title}({project_id}) のプロジェクトメンバから、{user_id} を脱退させました。")
            except requests.HTTPError as e:
                logger.warning(e)
                logger.warning(f"{project_title}({project_id}) のプロジェクトメンバから、{user_id} を脱退させられませんでした。")

    def drop_role_with_organization(self, organization_name: str, user_id_list: List[str]):
        projects = self.service.wrapper.get_all_projects_of_organization(
            organization_name, query_params={"account_id": self.service.api.account_id}
        )

        project_id_list = [e["project_id"] for e in projects]
        self.drop_role_with_project_id(project_id_list, user_id_list=user_id_list)

    def drop_role_with_project_id(self, project_id_list: List[str], user_id_list: List[str]):
        for project_id in project_id_list:
            try:
                if not self.facade.my_role_is_owner(project_id):
                    logger.warning(f"オーナではないため、プロジェクトメンバを脱退させられません。" f"project_id = {project_id}")
                    continue

                self.drop_project_members(project_id, user_id_list)

            except requests.HTTPError as e:
                logger.warning(e)
                logger.warning(f"project_id={project_id} のプロジェクトメンバからユーザを脱退させられませんでした。")


class DropProjectMembers(AbstractCommandLineInterface):
    """
    ユーザをプロジェクトから削除する
    """

    def main(self):
        args = self.args
        user_id_list = annofabcli.common.cli.get_list_from_args(args.user_id)

        main_obj = DropProjectMembersMain(self.service)
        if args.organization is not None:
            main_obj.drop_role_with_organization(args.organization, user_id_list)

        elif args.project_id is not None:
            project_id_list = annofabcli.common.cli.get_list_from_args(args.project_id)
            main_obj.drop_role_with_project_id(project_id_list, user_id_list)


def main(args):
    service = build_annofabapi_resource_and_login(args)
    facade = AnnofabApiFacade(service)
    DropProjectMembers(service, facade, args).main()


def parse_args(parser: argparse.ArgumentParser):
    parser.add_argument(
        "-u",
        "--user_id",
        type=str,
        nargs="+",
        required=True,
        help="脱退させるユーザのuser_idを指定してください。`file://`を先頭に付けると、一覧が記載されたファイルを指定できます。",
    )

    drop_group = parser.add_mutually_exclusive_group(required=True)
    drop_group.add_argument(
        "-p",
        "--project_id",
        type=str,
        nargs="+",
        help="脱退させるプロジェクトのproject_idを指定してください。`file://`を先頭に付けると、一覧が記載されたファイルを指定できます。",
    )

    drop_group.add_argument(
        "-org", "--organization", type=str, help="組織名を指定すると、組織配下のすべてのプロジェクト（自分が所属している）から、ユーザを脱退させます。"
    )

    parser.set_defaults(subcommand_func=main)


def add_parser(subparsers: argparse._SubParsersAction):
    subcommand_name = "delete"
    subcommand_help = "複数のプロジェクトから、ユーザを脱退させます。"
    description = "複数のプロジェクトから、ユーザを脱退させます。"
    epilog = "オーナロールを持つユーザで実行してください。"

    parser = annofabcli.common.cli.add_parser(subparsers, subcommand_name, subcommand_help, description, epilog=epilog)
    parse_args(parser)



import argparse
import logging
from typing import Dict, List, Optional, Union

import annofabapi
from annofabapi.models import ProjectMemberRole

import annofabcli
import annofabcli.common.cli
from annofabcli import AnnofabApiFacade
from annofabcli.common.cli import (
    AbstracCommandCinfirmInterface,
    AbstractCommandLineInterface,
    ArgumentParser,
    build_annofabapi_resource_and_login,
)

logger = logging.getLogger(__name__)

Metadata = Dict[str, Union[str, bool, int]]


class UpdateMetadataOfTaskMain(AbstracCommandCinfirmInterface):
    def __init__(self, service: annofabapi.Resource, all_yes: bool = False):
        self.service = service
        AbstracCommandCinfirmInterface.__init__(self, all_yes)

    def update_metadata_of_task(
        self, project_id: str, task_id_list: List[str], metadata: Metadata, batch_size: Optional[int] = None
    ):
        if batch_size is None:
            logger.info(f"{len(task_id_list)} 件のタスクのmetadataを、{metadata} に変更します。")
            request_body = {task_id: metadata for task_id in task_id_list}
            self.service.api.patch_tasks_metadata(project_id, request_body=request_body)

        else:
            logger.info(f"{len(task_id_list)} 件のタスクのmetadataを{metadata} に、{batch_size}個ずつ変更します。")
            first_index = 0
            while first_index < len(task_id_list):
                logger.info(
                    f"{first_index+1} 〜 {min(first_index+batch_size, len(task_id_list)-1)} 件目のタスクのmetadataを更新します。"
                )
                request_body = {task_id: metadata for task_id in task_id_list[first_index : first_index + batch_size]}
                self.service.api.patch_tasks_metadata(project_id, request_body=request_body)
                first_index += batch_size


class UpdateMetadataOfTask(AbstractCommandLineInterface):
    def main(self):
        args = self.args
        task_id_list = annofabcli.common.cli.get_list_from_args(args.task_id)
        metadata = annofabcli.common.cli.get_json_from_args(args.metadata)
        super().validate_project(args.project_id, [ProjectMemberRole.OWNER, ProjectMemberRole.TRAINING_DATA_USER])
        main_obj = UpdateMetadataOfTaskMain(self.service)
        main_obj.update_metadata_of_task(
            args.project_id, task_id_list=task_id_list, metadata=metadata, batch_size=args.batch_size
        )


def main(args):
    service = build_annofabapi_resource_and_login(args)
    facade = AnnofabApiFacade(service)
    UpdateMetadataOfTask(service, facade, args).main()


def parse_args(parser: argparse.ArgumentParser):
    argument_parser = ArgumentParser(parser)
    argument_parser.add_project_id()
    argument_parser.add_task_id(required=True)

    parser.add_argument(
        "--metadata",
        required=True,
        type=str,
        help="タスクに設定する`metadata`をJSON形式で指定してください。メタデータの値には文字列、数値、真偽値のいずれかを指定してください。"
        "`file://`を先頭に付けると、JSON形式のファイルを指定できます。",
    )

    parser.add_argument(
        "--batch_size",
        required=False,
        default=500,
        type=int,
        help="タスクのメタデータを何個ごとに更新するかを指定してください。一度に更新するタスクが多いとタイムアウトが発生する恐れがあります。",
    )

    parser.set_defaults(subcommand_func=main)


def add_parser(subparsers: argparse._SubParsersAction):
    subcommand_name = "update_metadata"
    subcommand_help = "タスクのメタデータを更新します。"
    description = "タスクのメタデータを上書きして更新します。"
    epilog = "オーナまたはアノテーションユーザロールを持つユーザで実行してください。"
    parser = annofabcli.common.cli.add_parser(
        subparsers, subcommand_name, subcommand_help, description=description, epilog=epilog
    )
    parse_args(parser)



import argparse
import logging
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Dict, List, Optional, Set

import numpy
import pandas

import annofabcli
from annofabcli import AnnofabApiFacade
from annofabcli.common.cli import AbstractCommandLineInterface, build_annofabapi_resource_and_login, get_list_from_args
from annofabcli.common.utils import print_csv, read_multiheader_csv
from annofabcli.statistics.csv import FILENAME_PEFORMANCE_PER_USER

logger = logging.getLogger(__name__)


@dataclass
class ResultDataframe:
    annotation_productivity: pandas.DataFrame
    inspection_acceptance_productivity: pandas.DataFrame
    quality_per_task: pandas.DataFrame
    quality_per_annotation: pandas.DataFrame


def join_annotation_productivity(
    df: pandas.DataFrame,
    df_performance: pandas.DataFrame,
    project_title: str,
    threshold_worktime: float,
    threshold_task_count: int,
) -> pandas.DataFrame:
    df_joined = df_performance
    df_joined = df_joined.set_index("user_id")

    df_joined = df_joined[
        (df_joined[("task_count", "annotation")] > threshold_task_count)
        | (df_joined[("prediction_actual_worktime_hour", "annotation")] >= threshold_worktime)
    ]
    df_tmp = df_joined[[("actual_worktime/annotation_count", "annotation")]]
    df_tmp.columns = pandas.MultiIndex.from_tuples([(project_title, "actual_worktime/annotation_count__annotation")])
    return df.join(df_tmp)


def join_inspection_acceptance_productivity(
    df: pandas.DataFrame,
    df_performance: pandas.DataFrame,
    project_title: str,
    threshold_worktime: float,
    threshold_task_count: int,
) -> pandas.DataFrame:
    def _join_inspection():
        if ("actual_worktime/annotation_count", "inspection") not in df_performance.columns:
            return df

        df_joined = df_performance[
            (df_performance[("task_count", "inspection")] > threshold_task_count)
            | (df_performance[("prediction_actual_worktime_hour", "inspection")] > threshold_worktime)
        ]
        df_joined.set_index("user_id", inplace=True)
        df_tmp = df_joined[[("actual_worktime/annotation_count", "inspection")]]
        df_tmp.columns = pandas.MultiIndex.from_tuples(
            [(project_title, "actual_worktime/annotation_count__inspection")]
        )
        return df.join(df_tmp)

    def _join_acceptance():
        if ("actual_worktime/annotation_count", "acceptance") not in df_performance.columns:
            return df

        df_joined = df_performance[
            (df_performance[("task_count", "acceptance")] > threshold_task_count)
            | (df_performance[("prediction_actual_worktime_hour", "acceptance")] > threshold_worktime)
        ]
        df_joined.set_index("user_id", inplace=True)
        df_tmp = df_joined[[("actual_worktime/annotation_count", "acceptance")]]
        df_tmp.columns = pandas.MultiIndex.from_tuples(
            [(project_title, "actual_worktime/annotation_count__acceptance")]
        )
        return df.join(df_tmp)

    df = _join_inspection()
    df = _join_acceptance()

    return df


def join_quality_per_task(
    df: pandas.DataFrame,
    df_performance: pandas.DataFrame,
    project_title: str,
    threshold_worktime: float,
    threshold_task_count: int,
) -> pandas.DataFrame:
    df_joined = df_performance
    df_joined = df_joined.set_index("user_id")

    df_joined = df_joined[
        (df_joined[("task_count", "annotation")] > threshold_task_count)
        | (df_joined[("prediction_actual_worktime_hour", "annotation")] > threshold_worktime)
    ]
    df_tmp = df_joined[[("rejected_count/task_count", "annotation")]]

    df_tmp.columns = pandas.MultiIndex.from_tuples([(project_title, "rejected_count/task_count")])
    return df.join(df_tmp)


def join_quality_per_annotation(
    df: pandas.DataFrame,
    df_performance: pandas.DataFrame,
    project_title: str,
    threshold_worktime: float,
    threshold_task_count: int,
) -> pandas.DataFrame:
    df_joined = df_performance
    df_joined = df_joined.set_index("user_id")

    df_joined = df_joined[
        (df_joined[("task_count", "annotation")] > threshold_task_count)
        | (df_joined[("prediction_actual_worktime_hour", "annotation")] > threshold_worktime)
    ]
    df_tmp = df_joined[[("pointed_out_inspection_comment_count/annotation_count", "annotation")]]

    df_tmp.columns = pandas.MultiIndex.from_tuples(
        [(project_title, "pointed_out_inspection_comment_count/annotation_count")]
    )
    return df.join(df_tmp)


def create_rating_df(
    df_user: pandas.DataFrame, target_dir: Path, threshold_worktime: float, threshold_task_count: int
) -> ResultDataframe:

    df_annotation_productivity = df_user
    df_inspection_acceptance_productivity = df_user
    df_quality_per_task = df_user
    df_quality_per_annotation = df_user

    for project_dir in target_dir.iterdir():
        if not project_dir.is_dir():
            continue

        csv = project_dir / FILENAME_PEFORMANCE_PER_USER
        project_title = project_dir.name
        if not csv.exists():
            logger.warning(f"{csv} は存在しないのでスキップします。")
            continue

        df_performance = read_multiheader_csv(str(csv), header_row_count=2)
        df_annotation_productivity = join_annotation_productivity(
            df_annotation_productivity,
            df_performance,
            project_title=project_title,
            threshold_worktime=threshold_worktime,
            threshold_task_count=threshold_task_count,
        )
        df_inspection_acceptance_productivity = join_inspection_acceptance_productivity(
            df_inspection_acceptance_productivity,
            df_performance,
            project_title=project_title,
            threshold_worktime=threshold_worktime,
            threshold_task_count=threshold_task_count,
        )
        df_quality_per_task = join_quality_per_task(
            df_quality_per_task,
            df_performance,
            project_title=project_title,
            threshold_worktime=threshold_worktime,
            threshold_task_count=threshold_task_count,
        )
        df_quality_per_annotation = join_quality_per_annotation(
            df_quality_per_annotation,
            df_performance,
            project_title=project_title,
            threshold_worktime=threshold_worktime,
            threshold_task_count=threshold_task_count,
        )

    return ResultDataframe(
        annotation_productivity=df_annotation_productivity.reset_index(),
        inspection_acceptance_productivity=df_inspection_acceptance_productivity.reset_index(),
        quality_per_task=df_quality_per_task.reset_index(),
        quality_per_annotation=df_quality_per_annotation.reset_index(),
    )


def output_csv(result: ResultDataframe, output_dir: Path):
    print_csv(result.annotation_productivity, str(output_dir / "annotation_productivity.csv"))
    print_csv(result.inspection_acceptance_productivity, str(output_dir / "inspection_acceptance_productivity.csv"))
    print_csv(result.quality_per_task, str(output_dir / "annotation_quality_per_task.csv"))
    print_csv(result.quality_per_annotation, str(output_dir / "annotation_quality_per_annotation.csv"))


def to_rank(series: pandas.Series) -> pandas.Series:
    quantile = list(series.quantile([0.25, 0.5, 0.75]))

    def _to_point(value):
        if numpy.isnan(value):
            return numpy.nan
        elif value < quantile[0]:
            return "A"
        elif quantile[0] <= value < quantile[1]:
            return "B"
        elif quantile[1] <= value < quantile[2]:
            return "C"
        elif quantile[2] <= value:
            return "D"
        else:
            return numpy.nan

    if len([v for v in series if not numpy.isnan(v)]) >= 4:
        return series.map(_to_point)
    else:
        return pandas.Series([numpy.nan] * len(series))


def to_deviation(series: pandas.Series, threshold_deviation_user_count: Optional[int] = None) -> pandas.Series:
    if series.count() == 0 or (
        threshold_deviation_user_count is not None and series.count() <= threshold_deviation_user_count
    ):
        return pandas.Series([numpy.nan] * len(series))
    else:
        std = series.std(ddof=0)
        mean = series.mean()
        if std != 0.0:
            return series.map(lambda x: (x - mean) / std * 10 + 50)
        else:
            return series.map(lambda x: 50 if not numpy.isnan(x) else numpy.nan)


def create_rank_df(df: pandas.DataFrame) -> pandas.DataFrame:
    df_rank = df.copy()
    for col in df.columns[3:]:
        df_rank[col] = to_rank(df[col])
    return df_rank


def create_deviation_df(
    df: pandas.DataFrame, threshold_deviation_user_count: Optional[int] = None, user_id_set: Optional[Set[str]] = None
) -> pandas.DataFrame:
    df_rank = df.copy()
    user_columns = df.columns[0:3]
    project_columns = df.columns[3:]
    for col in project_columns:
        df_rank[col] = to_deviation(df[col], threshold_deviation_user_count)

    df_rank["mean_of_deviation"] = df_rank[project_columns].mean(axis=1)
    df_rank["count_of_project"] = df_rank[project_columns].count(axis=1)
    df = df_rank[list(user_columns) + [("mean_of_deviation", ""), ("count_of_project", "")] + list(project_columns)]
    if user_id_set is not None:
        return df[df[("user_id", "")].isin(user_id_set)]
    else:
        return df


def create_basic_statistics_df(df: pandas.DataFrame) -> pandas.DataFrame:
    df_stat = df.describe().T
    return df_stat


def output_rank_csv(result: ResultDataframe, output_dir: Path):
    print_csv(create_rank_df(result.annotation_productivity), str(output_dir / "annotation_productivity_rank.csv"))
    print_csv(
        create_rank_df(result.inspection_acceptance_productivity),
        str(output_dir / "inspection_acceptance_productivity_rank.csv"),
    )
    print_csv(create_rank_df(result.quality_per_task), str(output_dir / "annotation_quality_per_task_rank.csv"))
    print_csv(
        create_rank_df(result.quality_per_annotation), str(output_dir / "annotation_quality_per_annotation_rank.csv")
    )


def output_deviation_csv(
    result: ResultDataframe,
    output_dir: Path,
    threshold_deviation_user_count: Optional[int] = None,
    user_id_list: Optional[List[str]] = None,
):
    """
    偏差値に変換したものを出力する
    """
    user_id_set: Optional[Set[str]] = set(user_id_list) if user_id_list is not None else None
    print_csv(
        create_deviation_df(result.annotation_productivity, threshold_deviation_user_count, user_id_set=user_id_set),
        str(output_dir / "annotation_productivity_deviation.csv"),
    )
    print_csv(
        create_deviation_df(
            result.inspection_acceptance_productivity, threshold_deviation_user_count, user_id_set=user_id_set
        ),
        str(output_dir / "inspection_acceptance_productivity_deviation.csv"),
    )
    print_csv(
        create_deviation_df(result.quality_per_task, threshold_deviation_user_count, user_id_set=user_id_set),
        str(output_dir / "annotation_quality_per_task_deviation.csv"),
    )

    print_csv(
        create_deviation_df(result.quality_per_annotation, threshold_deviation_user_count, user_id_set=user_id_set),
        str(output_dir / "annotation_quality_per_annotation_deviation.csv"),
    )


def output_basic_statistics_by_project(result: ResultDataframe, output_dir: Path):
    """
    プロジェクトごとの基本統計情報を出力する。
    """
    to_csv_kwargs = {"index": True}

    print_csv(
        create_basic_statistics_df(result.annotation_productivity),
        str(output_dir / "annotation_productivity_summary.csv"),
        to_csv_kwargs=to_csv_kwargs,
    )
    print_csv(
        create_basic_statistics_df(result.inspection_acceptance_productivity),
        str(output_dir / "inspection_acceptance_productivity_summary.csv"),
        to_csv_kwargs=to_csv_kwargs,
    )
    print_csv(
        create_basic_statistics_df(result.quality_per_task),
        str(output_dir / "annotation_quality_per_task_summary.csv"),
        to_csv_kwargs=to_csv_kwargs,
    )
    print_csv(
        create_basic_statistics_df(result.quality_per_annotation),
        str(output_dir / "annotation_quality_per_annotation_summary.csv"),
        to_csv_kwargs=to_csv_kwargs,
    )


def create_user_df(target_dir: Path) -> pandas.DataFrame:
    """
    "メンバごとの生産性と品質.csv"から、ユーザのDataFrameを作成する。

    Args:
        target_dir:

    Returns:
        ユーザのDataFrame. columnは("username", ""), ("biography", "") , indexが"user_id"

    """
    all_user_list: List[Dict[str, Any]] = []
    for project_dir in target_dir.iterdir():
        if not project_dir.is_dir():
            continue

        csv = project_dir / FILENAME_PEFORMANCE_PER_USER
        if not csv.exists():
            logger.warning(f"{csv} は存在しないのでスキップします。")
            continue

        tmp_df_user = read_multiheader_csv(str(csv), header_row_count=2)[
            [("user_id", ""), ("username", ""), ("biography", "")]
        ]
        all_user_list.extend(tmp_df_user.to_dict("records"))

    index = pandas.MultiIndex.from_tuples([("user_id", ""), ("username", ""), ("biography", "")])
    df_user = pandas.DataFrame(all_user_list, columns=index)
    df_user.drop_duplicates(inplace=True)
    return df_user.sort_values("user_id").set_index("user_id")


class WritePerformanceRatingCsv(AbstractCommandLineInterface):
    def main(self) -> None:
        args = self.args

        target_dir: Path = args.dir
        user_id_list = get_list_from_args(args.user_id) if args.user_id is not None else None
        df_user = create_user_df(target_dir)

        result = create_rating_df(
            df_user,
            target_dir,
            threshold_worktime=args.threshold_worktime,
            threshold_task_count=args.threshold_task_count,
        )

        output_dir: Path = args.output_dir
        output_csv(result, output_dir)
        output_rank_csv(result, output_dir)
        output_deviation_csv(
            result,
            output_dir,
            threshold_deviation_user_count=args.threshold_deviation_user_count,
            user_id_list=user_id_list,
        )
        output_basic_statistics_by_project(result, output_dir)


def parse_args(parser: argparse.ArgumentParser):
    parser.add_argument(
        "--dir",
        type=Path,
        required=True,
        help="プロジェクトディレクトリが存在するディレクトリを指定してください。プロジェクトディレクトリ内の`メンバごとの生産性と品質.csv`というファイルを読み込みます。",
    )

    parser.add_argument(
        "-u",
        "--user_id",
        type=str,
        nargs="+",
        help="評価対象のユーザのuser_idを指定してください。" "`file://`を先頭に付けると、user_idの一覧が記載されたファイルを指定できます。",
    )

    parser.add_argument(
        "--threshold_worktime",
        type=int,
        default=0,
        help="作業時間の閾値。この時間以下の作業者は除外する。`threshold_task_count`とはOR条件で絞り込まれる。",
    )
    parser.add_argument(
        "--threshold_task_count",
        type=int,
        default=0,
        help="作業したタスク数の閾値。このタスク数以下の作業者は除外する。`threshold_worktime`とはOR条件で絞り込まれる。",
    )
    parser.add_argument(
        "--threshold_deviation_user_count",
        type=int,
        help="偏差値を出す際、プロジェクト内の作業者がしきい値以下であれば、偏差値を算出しない。",
    )
    parser.add_argument("-o", "--output_dir", required=True, type=Path, help="出力ディレクトリ")

    parser.set_defaults(subcommand_func=main)


def main(args):
    service = build_annofabapi_resource_and_login(args)
    facade = AnnofabApiFacade(service)
    WritePerformanceRatingCsv(service, facade, args).main()


def add_parser(subparsers: argparse._SubParsersAction):
    subcommand_name = "write_performance_rating_csv"
    subcommand_help = "プロジェクトごとユーザごとにパフォーマンスを評価できる複数のCSVを出力します。"
    description = "プロジェクトごとユーザごとにパフォーマンスを評価できる複数のCSVを出力します。"
    parser = annofabcli.common.cli.add_parser(subparsers, subcommand_name, subcommand_help, description=description)
    parse_args(parser)



import argparse
import logging
import os
import shutil
import zipfile
from collections import defaultdict
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Dict, List, Optional, Set

from annofabapi.models import TaskStatus
from annofabapi.parser import SimpleAnnotationParser, lazy_parse_simple_annotation_dir, lazy_parse_simple_annotation_zip

import annofabcli
import annofabcli.common.cli
from annofabcli.common.facade import TaskQuery
from annofabcli.common.image import IsParserFunc

logger = logging.getLogger(__name__)


@dataclass
class FilterQuery:
    task_query: Optional[TaskQuery] = None
    task_id_set: Optional[Set[str]] = None
    exclude_task_id_set: Optional[Set[str]] = None
    input_data_id_set: Optional[Set[str]] = None
    exclude_input_data_id_set: Optional[Set[str]] = None
    input_data_name_set: Optional[Set[str]] = None
    exclude_input_data_name_set: Optional[Set[str]] = None


def _match_task_query(annotation: Dict[str, Any], task_query: Optional[TaskQuery]) -> bool:
    if task_query is None:
        return True

    if task_query.status is not None and annotation["task_status"] != task_query.status.value:
        return False

    if task_query.phase is not None and annotation["task_phase"] != task_query.phase:
        return False

    if task_query.phase_stage is not None and annotation["task_phase_stage"] != task_query.phase_stage:
        return False

    return True


def match_query(  # pylint: disable=too-many-return-statements
    annotation: Dict[str, Any], filter_query: FilterQuery
) -> bool:
    if filter_query.task_query is not None and not _match_task_query(annotation, filter_query.task_query):
        return False

    # xxx_set は1個のみ指定されていること前提なので、elif を羅列する
    if filter_query.task_id_set is not None and annotation["task_id"] not in filter_query.task_id_set:
        return False
    elif filter_query.exclude_task_id_set is not None and annotation["task_id"] in filter_query.exclude_task_id_set:
        return False
    elif (
        filter_query.input_data_id_set is not None and annotation["input_data_id"] not in filter_query.input_data_id_set
    ):
        return False
    elif (
        filter_query.exclude_input_data_id_set is not None
        and annotation["input_data_id"] in filter_query.exclude_input_data_id_set
    ):
        return False
    elif (
        filter_query.input_data_name_set is not None
        and annotation["input_data_name"] not in filter_query.input_data_name_set
    ):
        return False
    elif (
        filter_query.exclude_input_data_name_set is not None
        and annotation["input_data_name"] in filter_query.exclude_input_data_name_set
    ):
        return False

    return True


def create_outer_filepath_dict(namelist: List[str]) -> Dict[str, List[str]]:
    """
    外部アノテーションのファイルパス一覧

    Args:
        namelist: zipファイル内のファイル一覧

    Returns:

    """
    d = defaultdict(list)
    for name in namelist:
        tmp = name.split("/")
        if len(tmp) != 3:
            continue
        dirname = f"{tmp[0]}/{tmp[1]}"
        d[dirname].append(name)
    return d


class FilterAnnotation:
    @staticmethod
    def create_is_target_parser_func(
        task_status_complete: bool = False, task_id_list: Optional[List[str]] = None
    ) -> IsParserFunc:
        def is_target_parser(parser: SimpleAnnotationParser) -> bool:
            simple_annotation = parser.parse()
            if task_status_complete:
                if simple_annotation.task_status != TaskStatus.COMPLETE:
                    logger.debug(
                        f"task_statusがcompleteでない( {simple_annotation.task_status.value})ため、"
                        f"{simple_annotation.task_id}, {simple_annotation.input_data_name} はスキップします。"
                    )
                    return False

            if task_id_list is not None and len(task_id_list) > 0:
                if simple_annotation.task_id not in task_id_list:
                    logger.debug(
                        f"画像化対象外のタスク {simple_annotation.task_id} であるため、 {simple_annotation.input_data_name} はスキップします。"
                    )
                    return False

            return True

        return is_target_parser

    @staticmethod
    def filter_annotation_zip(annotation_zip: Path, filter_query: FilterQuery, output_dir: Path):
        with zipfile.ZipFile(str(annotation_zip)) as zip_file:
            zip_filepath_dict = create_outer_filepath_dict(zip_file.namelist())
            count = 0
            for parser in lazy_parse_simple_annotation_zip(annotation_zip):

                if not match_query(parser.load_json(), filter_query):
                    continue

                # JSONを展開
                zip_file.extract(parser.json_file_path, str(output_dir))
                # 塗りつぶしアノテーションが格納されているディレクトリを展開
                outer_annotation_dir = os.path.splitext(parser.json_file_path)[0]
                outer_annotation_file_list = zip_filepath_dict.get(outer_annotation_dir)
                if outer_annotation_file_list is not None:
                    for outer_annotation_file in outer_annotation_file_list:
                        zip_file.extract(outer_annotation_file, str(output_dir))
                count += 1
                if count % 10000 == 0:
                    logger.debug(f"{count} 件のJSONファイルとそれに紐づく塗りつぶし画像を {output_dir} に展開しました。")

            logger.info(f"{count} 件のJSONファイルとそれに紐づく塗りつぶし画像を {output_dir} に展開しました。")

    @staticmethod
    def filter_annotation_dir(annotation_dir: Path, filter_query: FilterQuery, output_dir: Path):
        count = 0
        for parser in lazy_parse_simple_annotation_dir(annotation_dir):
            if not match_query(parser.load_json(), filter_query):
                continue

            # JSONファイルをコピー
            dest_task_id_dir = output_dir / parser.task_id
            dest_task_id_dir.mkdir(exist_ok=True, parents=True)
            shutil.copy(str(annotation_dir / parser.json_file_path), str(dest_task_id_dir))
            # 塗りつぶしアノテーションファイルをコピー
            outer_annotation_dir = annotation_dir / os.path.splitext(parser.json_file_path)[0]
            if outer_annotation_dir.exists():
                shutil.copytree(str(outer_annotation_dir), str(output_dir))

            count += 1
            if count % 10000 == 0:
                logger.debug(f"{count} 件のJSONファイルとそれに紐づく塗りつぶし画像を {output_dir} をコピーしました。")

        logger.debug(f"{count} 件のJSONファイルとそれに紐づく塗りつぶし画像を {output_dir} をコピーしました。")

    @staticmethod
    def create_filter_query(args: argparse.Namespace) -> FilterQuery:
        task_query = (
            TaskQuery.from_dict(annofabcli.common.cli.get_json_from_args(args.task_query))
            if args.task_query is not None
            else None
        )

        task_id_set = set(annofabcli.common.cli.get_list_from_args(args.task_id)) if args.task_id is not None else None
        exclude_task_id_set = (
            set(annofabcli.common.cli.get_list_from_args(args.exclude_task_id))
            if args.exclude_task_id is not None
            else None
        )
        input_data_id_set = (
            set(annofabcli.common.cli.get_list_from_args(args.input_data_id))
            if args.input_data_id is not None
            else None
        )
        exclude_input_data_id_set = (
            set(annofabcli.common.cli.get_list_from_args(args.exclude_input_data_id))
            if args.exclude_input_data_id is not None
            else None
        )
        input_data_name_set = (
            set(annofabcli.common.cli.get_list_from_args(args.input_data_name))
            if args.input_data_name is not None
            else None
        )
        exclude_input_data_name_set = (
            set(annofabcli.common.cli.get_list_from_args(args.exclude_input_data_name))
            if args.exclude_input_data_name is not None
            else None
        )

        return FilterQuery(
            task_query=task_query,
            task_id_set=task_id_set,
            exclude_task_id_set=exclude_task_id_set,
            input_data_id_set=input_data_id_set,
            exclude_input_data_id_set=exclude_input_data_id_set,
            input_data_name_set=input_data_name_set,
            exclude_input_data_name_set=exclude_input_data_name_set,
        )

    def main(self, args: argparse.Namespace):
        logger.info(f"args: {args}")

        annotation_path: Path = args.annotation
        output_dir: Path = args.output_dir
        output_dir.mkdir(exist_ok=True, parents=True)
        filter_query = self.create_filter_query(args)

        if zipfile.is_zipfile(annotation_path):
            self.filter_annotation_zip(annotation_path, filter_query=filter_query, output_dir=output_dir)
        elif annotation_path.is_dir():
            self.filter_annotation_dir(annotation_path, filter_query=filter_query, output_dir=output_dir)
        else:
            logger.info("TODO")
            return


def main(args):
    FilterAnnotation().main(args)


def parse_args(parser: argparse.ArgumentParser):

    parser.add_argument("--annotation", type=Path, required=True, help="アノテーションzip、またはzipを展開したディレクトリ")

    parser.add_argument(
        "-tq",
        "--task_query",
        type=str,
        help="タスクを絞り込むためのクエリ条件をJSON形式で指定します。使用できるキーは status, phase, phase_stage です。"
        "`file://`を先頭に付けると、JSON形式のファイルを指定できます。",
    )

    id_name_list_group = parser.add_mutually_exclusive_group()
    id_name_list_group.add_argument(
        "-t",
        "--task_id",
        type=str,
        nargs="+",
        help="抽出するタスクのtask_idを指定してください。" + "`file://`を先頭に付けると、task_id の一覧が記載されたファイルを指定できます。",
    )

    id_name_list_group.add_argument(
        "--exclude_task_id",
        type=str,
        nargs="+",
        help="除外するタスクのtask_idを指定してください。" + "`file://`を先頭に付けると、task_id の一覧が記載されたファイルを指定できます。",
    )

    id_name_list_group.add_argument(
        "-i",
        "--input_data_id",
        type=str,
        nargs="+",
        help="抽出する入力データのinput_data_idを指定してください。" + "`file://`を先頭に付けると、input_data_id の一覧が記載されたファイルを指定できます。",
    )
    id_name_list_group.add_argument(
        "--exclude_input_data_id",
        type=str,
        nargs="+",
        help="除外する入力データのinput_data_idを指定してください。" + "`file://`を先頭に付けると、input_data_id の一覧が記載されたファイルを指定できます。",
    )

    id_name_list_group.add_argument(
        "--input_data_name",
        type=str,
        nargs="+",
        help="抽出する入力データのinput_data_nameを指定してください。" + "`file://`を先頭に付けると、input_data_name の一覧が記載されたファイルを指定できます。",
    )
    id_name_list_group.add_argument(
        "--exclude_input_data_name",
        type=str,
        nargs="+",
        help="除外する入力データのinput_data_nameを指定してください。" + "`file://`を先頭に付けると、input_data_name の一覧が記載されたファイルを指定できます。",
    )

    parser.add_argument("-o", "--output_dir", type=Path, required=True, help="出力先ディレクトリのパス")

    parser.set_defaults(subcommand_func=main)


def add_parser(subparsers: argparse._SubParsersAction):
    subcommand_name = "filter_annotation"

    subcommand_help = "アノテーションzipから特定のファイルを絞り込んで、zip展開します。"

    description = "アノテーションzipから特定のファイルを絞り込んで、zip展開します。"

    parser = annofabcli.common.cli.add_parser(subparsers, subcommand_name, subcommand_help, description)
    parse_args(parser)



import argparse

import annofabcli
import annofabcli.common.cli
import annofabcli.filesystem.filter_annotation
import annofabcli.filesystem.write_annotation_image


def parse_args(parser: argparse.ArgumentParser):

    subparsers = parser.add_subparsers()

    # サブコマンドの定義
    annofabcli.filesystem.filter_annotation.add_parser(subparsers)
    annofabcli.filesystem.write_annotation_image.add_parser(subparsers)


def add_parser(subparsers: argparse._SubParsersAction):
    subcommand_name = "filesystem"
    subcommand_help = "ファイル操作関係（Web APIにアクセスしない）のサブコマンド"
    description = "ファイル操作関係（Web APIにアクセスしない）のサブコマンド"

    parser = annofabcli.common.cli.add_parser(
        subparsers, subcommand_name, subcommand_help, description, is_subcommand=False
    )
    parse_args(parser)
